#!/usr/bin/env python3
# coding: utf-8

import argparse
import os
import re
import sys
from volkanic.utils import desktop_open

from joker.xopen import utils


class Client(object):
    def __init__(self, port=None):
        self.port = port or utils.get_port()

    def chksvr(self):
        return self.request('version').startswith(b'joker-xopen')

    def request(self, qs):
        if isinstance(qs, str):
            qs = qs.encode('utf-8')
        elif isinstance(qs, bytes):
            pass
        else:
            qs = str(qs).encode('utf-8')
        return utils.netcat('127.0.0.1', self.port, qs)

    def get(self, key):
        return self.request('get ' + key).decode()

    def open(self, key):
        location = self.get(key)
        if location:
            return desktop_open(location)

    def open_many(self, *keys):
        if not keys:
            return
        if not self.chksvr():
            return
        if len(keys) == 1:
            return self.open(keys[0])

        from concurrent.futures import ThreadPoolExecutor
        pool = ThreadPoolExecutor(max_workers=4)
        return pool.map(self.open, keys)


def amphib_open(*locators):
    if not locators:
        return
    direct_locators = set()
    indirect_locators = set()
    exists = os.path.exists

    for loc in locators:
        if loc.startswith('~'):
            path = os.path.expanduser(loc)
            if os.path.exists(path):
                direct_locators.add(path)
                continue
        if exists(loc) or re.match(r'(https?|file|ftp)://', loc):
            direct_locators.add(loc)
        elif re.match(r'\S{1,64}$', loc):
            indirect_locators.add(loc)
    desktop_open(*direct_locators)
    Client().open_many(*indirect_locators)


def get_and_print(*keys):
    client = Client()
    for k in keys:
        v = client.get(k)
        print(v)


def request_and_printerr(qs):
    Client().request(qs)
    msg = "xopen client: '{}' request sent".format(qs)
    print(msg, file=sys.stderr)


def runxopen(prog=None, args=None):
    if not prog and sys.argv[0].endswith('__main__.py'):
        prog = 'python3 -m joker.xopen'
    desc = 'joker-xopen client'
    pr = argparse.ArgumentParser(prog=prog, description=desc)
    add = pr.add_argument

    add('-g', '--get', action='store_true',
        help='get value from cache server')

    add('-r', '--reload', action='store_true',
        help='request server to reload from conf')

    add('-u', '--update', action='store_true',
        help='request server to update from conf')

    add('locators', metavar='locator', nargs='*',
        help='key (in conf), pathname or url')

    ns = pr.parse_args(args)

    default_locators = ['.']
    if ns.reload:
        request_and_printerr('reload')
        default_locators = []

    if ns.update:
        request_and_printerr('update')
        default_locators = []

    locators = ns.locators or default_locators
    if ns.get:
        return get_and_print(*ns.locators)
    amphib_open(*locators)



#!/usr/bin/env python3
# coding: utf-8

import os
import threading
from collections import OrderedDict


class LimitedDict(object):
    default_capacity = 2 ** 26

    def __init__(self, capacity=0, data=None):
        if not isinstance(data, OrderedDict):
            data = OrderedDict(data or {})
        self.data = data
        self.capacity = capacity or self.default_capacity
        self.freespace = capacity - self._estimate_size(self.data)
        self.evict()

    @staticmethod
    def _estimate_size(data):
        lengths = ((len(k) + len(v)) for k, v in data.items())
        return sum(lengths)

    def _reset(self, data):
        if not isinstance(data, OrderedDict):
            data = OrderedDict(data)
        with threading.Lock():
            self.data = data
            self.freespace = self.capacity - self._estimate_size(data)
            self.evict()

    def _update(self, data):
        with threading.Lock():
            self.data.update(data)
            for key in data:
                self.data.move_to_end(key, last=True)

    def pop(self, key, *default):
        val = self.data.pop(key, *default)
        if (val,) != default:
            self.freespace += len(key) + len(val)
        return val

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, val):
        pval = self.data.get(key, '')
        with threading.Lock():
            self.data[key] = val
            self.data.move_to_end(key, last=True)
            self.freespace += len(pval) - len(val)

    def get(self, key, default=None):
        return self.data.get(key, default)

    def evict(self):
        while self.freespace <= 0:
            self.data.popitem(last=False)


class ReconfDict(LimitedDict):
    comment = b'#'

    def __init__(self, capacity, *paths):
        _paths = {}
        for path in paths:
            path = os.path.expanduser(path)
            _paths[path] = self._getmtime(path)
        self.paths = _paths
        data = self.parse_many(*paths)
        LimitedDict.__init__(self, capacity, data)

    def __bool__(self):
        return True

    @staticmethod
    def _getmtime(path):
        return int(os.path.getmtime(path) * 1000)

    @classmethod
    def parse(cls, path):
        data = OrderedDict()
        with open(path, 'rb') as fin:
            for line in fin:
                line = line.strip()
                if not line or line.startswith(cls.comment):
                    continue
                parts = line.split(maxsplit=1)
                if len(parts) != 2:
                    continue
                k, v = parts
                data[k] = v
        return data

    @classmethod
    def parse_many(cls, *paths):
        data = OrderedDict()
        for path in paths:
            try:
                data.update(cls.parse(path))
            except (FileNotFoundError, IOError):
                continue
        return data

    def reload(self):
        updated = True
        for path, mtime in self.paths.items():
            real_mtime = self._getmtime(path)
            if real_mtime > mtime:
                updated = False
                break
        if not updated:
            data = self.parse_many(*self.paths)
            self._reset(data)

    def update(self):
        outdated = []
        for path, mtime in self.paths.items():
            real_mtime = self._getmtime(path)
            if real_mtime > mtime:
                outdated.append(path)
        if outdated:
            data = self.parse_many(*outdated)
            self._update(data)




#!/usr/bin/env python3
# coding: utf-8
import argparse
import os
import sys
import threading
import traceback
from collections import deque
from concurrent.futures import ThreadPoolExecutor

import requests
from joker.cast.syntax import printerr
from joker.minions.cache import CacheServer

from joker.xopen import utils
from joker.xopen.datatypes import LimitedDict, ReconfDict
from joker.xopen.environ import JokerInterface

ji = JokerInterface()


def _printerr(*args, **kwargs):
    parts = []
    for a in args:
        if isinstance(a, bytes):
            parts.append(a.decode())
        elif isinstance(a, (deque, list, tuple)):
            parts.extend(a)
    kwargs.setdefault('sep', ':')
    printerr(*parts, **kwargs)


def under_joker_xopen_dir(*paths):
    return ji.under_joker_subdir(*paths)


def get_default_source_paths():
    import glob
    xopen_dirpath = ji.under_joker_subdir(mkdirs=True)
    path = os.path.join(xopen_dirpath, 'xopen.txt')
    if not os.path.exists(path):
        with open(path, 'a'):
            pass
    patt = os.path.join(xopen_dirpath, '*.txt')
    return glob.glob(patt)


class _ReconfDict(ReconfDict):
    @classmethod
    def parse(cls, path):
        data = ReconfDict.parse(path)
        f = os.path.expanduser
        return {k: f(v) for k, v in data.items()}


class XopenCacheServer(CacheServer):
    def __init__(self, capacity, *paths):
        CacheServer.__init__(self)
        self.data = _ReconfDict(capacity, *paths)
        self.cache = LimitedDict(capacity, )
        self.cached_verbs = {b'http-get'}
        self.verbs = {
            b'reload': self.exec_reload,
            b'update': self.exec_update,
            b'version': self.exec_version,
            b'http-get': self.exec_http_get,
        }
        self._tpexec = ThreadPoolExecutor(max_workers=3)

    def _execute(self, verb, payload):
        try:
            return self.verbs[verb](payload)
        except Exception:
            traceback.print_exc()

    def _execute_with_cache(self, verb, payload):
        key = verb + b':' + payload
        try:
            return self.cache[key]
        except KeyError:
            pass
        rv = self._execute(verb, payload)
        self.cache[key] = rv
        return rv

    def execute(self, verb, payload):
        if verb in self.cached_verbs:
            return self._execute_with_cache(verb, payload)
        if verb in self.verbs:
            return self._execute(verb, payload)
        return CacheServer.execute(self, verb, payload)

    def _printdiff(self, vdata):
        udata = self.data.data
        keys = set(vdata)
        keys.update(udata)
        for k in keys:
            u = udata.get(k)
            v = vdata.get(k)
            if u != v:
                _printerr(k, u, v)

    @staticmethod
    def exec_http_get(url):
        return requests.get(url).content

    @staticmethod
    def exec_version(_):
        import joker.xopen
        return 'joker-xopen==' + joker.xopen.__version__

    def exec_reload(self, _):
        self._tpexec.submit(self.data.reload)

    def exec_update(self, _):
        self._tpexec.submit(self.data.update)

    def eviction(self, period=5):
        import time
        while True:
            time.sleep(period)
            self.cache.evict()
            self.data.evict()


def run(prog, args):
    import sys
    if not prog and sys.argv[0].endswith('server.py'):
        prog = 'python3 -m joker.xopen.server'
    desc = 'joker-xopen cache server'
    pr = argparse.ArgumentParser(prog=prog, description=desc)
    add = pr.add_argument

    add('-c', '--capacity', metavar='int',
        type=int, default=ReconfDict.default_capacity)

    add('sources', metavar='path', nargs='*',
        help='path to a source file, default `~/.joker/xopen/*.txt`')

    ns = pr.parse_args(args)
    sources = ns.sources or get_default_source_paths()

    # noinspection PyBroadException
    try:
        svr = XopenCacheServer(ns.capacity, *sources)
    except Exception:
        traceback.print_exc()
        sys.exit(1)
    threading.Thread(target=svr.eviction, daemon=True).start()
    svr.runserver('127.0.0.1', utils.get_port())


if __name__ == '__main__':
    run(None, sys.argv[1:])



#!/usr/bin/env python3
# coding: utf-8

import os


def get_port():
    envvar = 'JOKER_XOPEN_PORT'
    try:
        return int(os.environ.get(envvar))
    except TypeError:
        return 18831


def netcat(host, port, content):
    # https://stackoverflow.com/a/1909355/2925169
    import socket
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.connect((host, port))
        sock.sendall(content)
        sock.shutdown(socket.SHUT_WR)
        return sock.recv(2 ** 16)
    except Exception:
        return b''
    finally:
        sock.close()

